@@ -167,31 +167,39 @@ class NumerousLogHandler:
     #TODO make a handler thread to schedule update to server!
 
 class NumerousClient:
-    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, secure=None, token=None, clear_data=None, no_log=False, instance_id=None):
-
+    def __init__(self, job_id=None, project=None, scenario=None, server=None, port=None, refresh_token=None, clear_data=None, no_log=False, instance_id=None, secure=None):
         self.job_states = JobStatus
         self._project = get_env(project, 'NUMEROUS_PROJECT')
         self._scenario = get_env(scenario, 'NUMEROUS_SCENARIO')
-        self._token = get_env(token, 'NUMEROUS_API_TOKEN')
+        self._spm_job_id = get_env(job_id, 'SPM_JOB_ID')
+        self._refresh_token = get_env(refresh_token, 'NUMEROUS_API_REFRESH_TOKEN')
+        self._access_token = None
 
-        self._job_id = get_env(job_id, 'JOB_ID')
+        self._instance_id = str(uuid4()) if instance_id is None else instance_id
         self._execution_id = get_env(None, 'NUMEROUS_EXECUTION_ID', 'not_found')
         if self._execution_id == 'not_found' or self._execution_id is None:
             self._execution_id = str(uuid4())
+            print("EX", self._execution_id)
             log.debug(f'No execution id found - generating new: {self._execution_id}')
 
         self._complete = True
         self.writers = []
-        channel = self._init_channel(server=server, port=port, secure=secure, instance_id=instance_id)
+
+
+        channel = self._init_channel(server=server, port=port, secure=secure, instance_id=self._instance_id)
         stub = spm_pb2_grpc.SPMStub(channel)
 
-        # TODO: Maybe reconsider
+        # TODO: Maybe reconsider both of these!
         self._job_manager = spm_pb2_grpc.JobManagerStub(channel)
-
+        self._token_manager = spm_pb2_grpc.TokenManagerStub(channel)
         self._base_client = stub
 
-        # Register execution
-        self._base_client.RegisterExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._job_id, execution_id=self._execution_id))
+        # Refresh token every 9 minutes (?)
+        self._access_token_refresher = RepeatedFunction(
+            interval=9*60, function=self._refresh_access_token, run_initially=True,
+            refresh_token=self._refresh_token, instance_id=self._instance_id, execution_id=self._execution_id
+        )
+        self._access_token_refresher.start()
 
         self._terminated = False
         if not no_log:
@@ -207,18 +215,14 @@ class NumerousClient:
         clear_data = get_env(clear_data, 'NUMEROUS_CLEAR_DATA', default='False') == 'True'
 
         if clear_data:
-
             self.clear_timeseries_data()
-        # Initialize probe that updates timestamp every (300) seconds
-        # Uncomment to enable timed probe feature...
-        self.probe_doc_id = None
 
         self._listen = threading.Thread(target=self._listen_terminate)
         self._listen.setDaemon(True)
         self._listen.start()
 
     def complete_execution(self):
-        self._base_client.CompleteExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._job_id, execution_id=self._execution_id))
+        self._base_client.CompleteExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._spm_job_id, execution_id=self._execution_id))
 
     def _init_channel(self, server, port, secure=None, instance_id=None):
         secure_channel = get_env(secure, 'SECURE_CHANNEL')
@@ -232,15 +236,17 @@ class NumerousClient:
             with open(f'{os.path.dirname(os.path.abspath(__file__))}/certs/server.crt', 'rb') as f:
                 creds = grpc.ssl_channel_credentials(f.read())
             channel = grpc.secure_channel(f'{server}:{port}', creds)
-
         else:
             channel = grpc.insecure_channel(f'{server}:{port}')
 
-        vi = ValidationInterceptor(token=self._token, instance=instance_id)
+        vi = ValidationInterceptor(token=self._access_token, token_callback=self._get_current_token, instance=instance_id)
         self._instance = vi.instance
         channel = grpc.intercept_channel(channel, vi)
         return channel
 
+    def _get_current_token(self):
+        return self._access_token
+
     def _init_logger(self):
         log_queue = Queue(-1)
         numerous_logger = logging.handlers.QueueHandler(queue=log_queue)
@@ -251,10 +257,18 @@ class NumerousClient:
         self.log_listener = logging.handlers.QueueListener(log_queue, NumerousLogHandler(self))
         self.log_listener.start()
 
+    def _refresh_access_token(self, refresh_token, instance_id, execution_id):
+        print("REFRESH TOKEN", refresh_token)
+        token = self._token_manager.GetAccessToken(
+            spm_pb2.RefreshRequest(
+                refresh_token=spm_pb2.Token(val=refresh_token), instance_id=instance_id, execution_id=execution_id
+            )
+        )
+        print(f"TOKEN REFRESHED: {token.val}")
+        self._access_token = token.val
 
     def _listen_terminate(self):
-
-       for channel, message in self.subscribe_messages([".".join(['COMMAND', self._project, self._scenario, self._job_id])]):
+       for channel, message in self.subscribe_messages([".".join(['COMMAND', self._project, self._scenario, self._spm_job_id])]):
             log.debug('Received message: '+str(message)+'\n on channel: '+str(channel))
             if 'command' in message:
                 command = message['command']
@@ -275,7 +289,7 @@ class NumerousClient:
         if scenario_doc is None:
             scenario_doc, files = self.get_scenario_document()
 
-        return scenario_doc['jobs'][self._job_id]
+        return scenario_doc['jobs'][self._spm_job_id]
 
 
     def get_group_document(self, group):
@@ -349,7 +363,7 @@ class NumerousClient:
 
             self.last_progress=tic
             self._base_client.SetScenarioProgress(spm_pb2.ScenarioProgress(
-                project=self._project, scenario=scenario, spm_job_id=self._job_id,
+                project=self._project, scenario=scenario, spm_job_id=self._spm_job_id,
                 message=message, status=status, clean=clean, progress=progress
             ))
 
@@ -377,7 +391,7 @@ class NumerousClient:
         if scenario is None:
             scenario = self._scenario
 
-        self._base_client.PushScenarioError(spm_pb2.ScenarioError(project=self._project, scenario=scenario, error=error, spm_job_id=self._job_id))
+        self._base_client.PushScenarioError(spm_pb2.ScenarioError(project=self._project, scenario=scenario, error=error, spm_job_id=self._spm_job_id))
 
     def push_scenario_log(self, message, initialize=False, scenario=None):
         if scenario is None:
@@ -385,7 +399,14 @@ class NumerousClient:
 
         self._base_client.PushScenarioLogEntry(spm_pb2.ScenarioLogEntry(
             project=self._project, scenario=scenario, message=message,
-            spm_job_id=self._job_id, initialize=initialize
+            spm_job_id=self._spm_job_id, initialize=initialize
+        ))
+
+    def push_scenario_logs(self, logs):
+
+        self._base_client.PushExecutionLogEntries(spm_pb2.LogEntries(
+            execution_id=self._execution_id, log_entries=[l[1] for l in logs],
+            timestamps=[l[0].timestamp() for l in logs]
         ))
 
     def delete_scenario(self, scenario=None):
@@ -412,14 +433,14 @@ class NumerousClient:
             files_out[f.name] = local_file_path
         return files_out
 
-    def upload_file(self, local_file, file=None, scenario=None):
+    def upload_file(self, local_file, file_id, file=None, scenario=None):
         if scenario is None:
             scenario = self._scenario
 
         if file is None:
             file = local_file.split('/')[-1]
         print('File: ', file)
-        upload_url = self._base_client.GenerateScenarioUploadSignedURL(spm_pb2.ScenarioFilePath(project=self._project, scenario=scenario, path=file))
+        upload_url = self._base_client.GenerateScenarioUploadSignedURL(spm_pb2.ScenarioFilePath(project=self._project, scenario=scenario, path=file, file_id=file_id))
 
         def upload():
             filename = local_file
@@ -537,7 +558,7 @@ class NumerousClient:
             for r in l.data:
                 yield r.tag, r.values
 
-    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, subscribe=False):
+    def read_data_as_row(self, tags=[], start:int=0, end:int=0, scenario=None, subscribe=False, offset=0):
 
         log.debug('Reading data as row')
         log.debug('Subscribed: ' + str(subscribe) + ', starting from: '+str(start))
@@ -571,12 +592,17 @@ class NumerousClient:
 
                     if current_ix >= start and current_ix < end:
 
-                        yield {k: v[i] for k,v in current_row.items()}
+                        r = {k: v[i] for k,v in current_row.items()}
+                        r['_index']+=offset
+                        yield r
+
                 current_row = {}
 
         if len(list(current_row.keys()))>0:
             for i, ix in enumerate(current_row['_index']):
-                yield {k: v[i] for k, v in current_row.items()}
+                r = {k: v[i] for k, v in current_row.items()}
+                r['_index'] += offset
+                yield r
 
 
         log.debug('Reading as row complete')
@@ -714,6 +740,7 @@ class NumerousClient:
         static_data = {}
         input_source_types = {}
         only_static = True
+
         for c in scenario_data['simComponents']:
             if "subcomponents" in c:
                 for sc in c["subcomponents"]:
@@ -785,11 +812,15 @@ class NumerousClient:
         for message in self._base_client.SubscribeForUpdates(spm_pb2.Subscription(channel_patterns=channel_patterns)):
             yield message.channel, json.loads(message.message)
 
+    def refresh_token(self, refresh_token):
+        refreshed_token = self._token_manager.GetAccessToken(spm_pb2.Token(val=refresh_token))
+        return refreshed_token
 
     def close(self):
         #close all writers
         for w in self.writers:
             w.close()
+        self._access_token_refresher.stop()
 
         if hasattr(self,'log_listener'):
             self.log_listener.stop()
@@ -798,8 +829,8 @@ class NumerousClient:
 
 
 @contextmanager
-def open_client(job_id=None, project=None, scenario=None, clear_data=None, no_log=False) -> NumerousClient:
-    numerous_client_ = NumerousClient(job_id, project, scenario, clear_data=clear_data, no_log=no_log)
+def open_client(job_id=None, project=None, scenario=None, clear_data=None, no_log=False, instance_id=None, refresh_token=None) -> NumerousClient:
+    numerous_client_ = NumerousClient(job_id, project, scenario, clear_data=clear_data, no_log=no_log, instance_id=instance_id, refresh_token=refresh_token)
 
     try:
         yield numerous_client_
@@ -850,6 +881,7 @@ class DynamicSource:
         self.row_queue = Queue(100)
 
         def _next():
+            offset = 0
             while True:
                 if not self.row_queue.full():
                     first = True
@@ -858,7 +890,7 @@ class DynamicSource:
                         try:
                             row_ = self.generator.__next__()
                             #print(row_['_index'])
-
+                            offset = row_['_index']
                             self.row_queue.put(row_)
 
                         except StopIteration:
@@ -870,7 +902,7 @@ class DynamicSource:
                                 log.debug('Requerying data')
                                 self.generator = client.read_data_as_row(tags=[s['tag'] for s in self.spec.values()],
                                                                          scenario=scenario, start=0,
-                                                                         subscribe=subscribe)
+                                                                         subscribe=subscribe, offset=offset)
 
                 else:
                     sleep(.01)
@@ -960,16 +992,33 @@ class RepeatedFunction:
         self._timer.cancel()
         self.is_running = False
 
+
 class NumerousAdminClient(NumerousClient):
-    def __init__(self, token=None, server=None, port=None):
+    def __init__(self, server=None, port=None, secure=None, refresh_token=None, instance_id=None):
         self.writers = []
         self._user = 'test_user'
         self._organization = 'EnergyMachines'
+        self._execution_id = get_env(None, 'NUMEROUS_EXECUTION_ID', 'not_found')
 
-        self._token = get_env(token, 'NUMEROUS_API_TOKEN')
+#        self._token = get_env(token, 'NUMEROUS_API_TOKEN')
+        self._access_token = None
+        self._refresh_token = get_env(refresh_token, 'NUMEROUS_API_REFRESH_TOKEN')
+        self._instance_id = str(uuid4()) if instance_id is None else instance_id
+        secure = get_env(secure, 'SECURE_CHANNEL')
 
-        channel = self._init_channel(server=server, port=port)
+        channel = self._init_channel(server=server, port=port, secure=secure)
         stub = spm_pb2_grpc.SPMStub(channel)
+        self._token_manager = spm_pb2_grpc.TokenManagerStub(channel)
+
+        # Refresh token every 9 minutes (?)
+        self._access_token_refresher = RepeatedFunction(
+            interval=9 * 60, function=self._refresh_access_token, run_initially=True,
+            refresh_token=self._refresh_token, instance_id=self._instance_id, execution_id=self._execution_id
+        )
+        self._access_token_refresher.start()
+
+
+
 
         # TODO: Maybe reconsider
         self._job_manager = spm_pb2_grpc.JobManagerStub(channel)
@@ -978,6 +1027,11 @@ class NumerousAdminClient(NumerousClient):
         self._terminated = False
         self._complete = False
 
+        self._execution_id = get_env(None, 'NUMEROUS_EXECUTION_ID', 'not_found')
+        if self._execution_id == 'not_found' or self._execution_id is None:
+            self._execution_id = str(uuid4())
+            log.debug(f'No execution id found - generating new: {self._execution_id}')
+
         log.info('init admin client')
 
     """
@@ -1035,9 +1089,13 @@ class NumerousAdminClient(NumerousClient):
 
         return spm_pb2.Empty()
 
+    def read_execution_logs(self):
+        for l in self._base_client.ReadExecutionLogEntries(spm_pb2.ExecutionReadLogs(execution_id=self._execution_id, start=0.0, end=0.0)):
+            yield l.log_entry, datetime.fromtimestamp(l.timestamp)
+
 @contextmanager
-def open_admin_client() -> NumerousAdminClient:
-    numerous_client_ = NumerousAdminClient()
+def open_admin_client(server=None, port=None, secure=None, refresh_token=None) -> NumerousAdminClient:
+    numerous_client_ = NumerousAdminClient(server=server, port=port, secure=secure, refresh_token=refresh_token)
 
     try:
         yield numerous_client_
